/*
Copyright 2019 Benji Dial

Permission to use, copy, modify, and/or distribute this
software for any purpose with or without fee is hereby
granted, provided that the above copyright notice and this
permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
OF THIS SOFTWARE.
*/

#include "proc.h"
#include "mem.h"
#include "files.h"

uint16_t new_proc(uint8_t *file) {
  uint16_t handle = open_file(file);
  if (handle)
    for (uint16_t id = 3; id; ++id)
      if (!(PROCS[id].memory_start)) {
        uint32_t size = get_size(id);
        void *mem = PROCS[id].memory_start = allocate_pages(PROCS[id].n_pages = (size - 1) / 4096 + 1, id);
        read_file(handle, size, mem);
        close_file(handle);
        //Process file header and make new process
        return id;
      }
  close_file(handle);
  return 0;
}

void end_proc(uint16_t id) {
  deallocate_pages(PROCS[id].memory_start, PROCS[id].n_pages);
  PROCS[id].memory_start = 0;
}